from collections import namedtuple
from distutils.version import LooseVersion
from graphviz import Digraph
import torch
from torch.autograd import Variable
#from stlcg import STL_Formula
import IPython
from stl import And, Equal, GreaterThan, LessThan, Negation, Or, convert_to_input_values

Node = namedtuple('Node', ('name', 'inputs', 'attr', 'op'))

class STL_Formula(torch.nn.Module):
    
    """MIT License

    Copyright (c) 2019 Karen Leung, Nikos Arechiga"""
    '''
    NOTE: All the inputs are assumed to be TIME REVERSED. The outputs are also TIME REVERSED
    All STL formulas have the following functions:
    robustness_trace: Computes the robustness trace.
    robustness: Computes the robustness value of the trace
    eval_trace: Computes the robustness trace and returns True in each entry if the robustness value is > 0
    eval: Computes the robustness value and returns True if the robustness value is > 0
    forward: The forward function of this STL_formula PyTorch module (default to the robustness_trace function)

    Inputs to these functions:
    trace: the input signal assumed to be TIME REVERSED. If the formula has two subformulas (e.g., And), then it is a tuple of the two inputs. An input can be a tensor of size [batch_size, time_dim,...], or an Expression with a .value (Tensor) associated with the expression.
    pscale: predicate scale. Default: 1
    scale: scale for the max/min function.  Default: -1
    keepdim: Output shape is the same as the input tensor shapes. Default: True
    agm: Use arithmetic-geometric mean. (In progress.) Default: False
    distributed: Use the distributed mean. Default: False
    '''

    def __init__(self):
        super(STL_Formula, self).__init__()

    def robustness_trace(self, trace, pscale=1, scale=-1, keepdim=True, agm=False, distributed=False, **kwargs):
        raise NotImplementedError("robustness_trace not yet implemented")

    def robustness(self, inputs, time=0, pscale=1, scale=-1, keepdim=True, agm=False,distributed=False, **kwargs):
        '''
        Extracts the robustness_trace value at the given time.
        Default: time=0 assuming this is the index for the NON-REVERSED trace. But the code will take it from the end since the input signal is TIME REVERSED.

        '''
        return self.forward(inputs, pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs)[:,-(time+1),:].unsqueeze(1)

    def eval_trace(self, inputs, pscale=1, scale=-1, keepdim=True, agm=False, distributed=False, **kwargs):
        '''
        The values in eval_trace are 0 or 1 (False or True)
        '''
        return self.forward(inputs, pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs) > 0

    def eval(self, inputs, time=0, pscale=1, scale=-1, keepdim=True, agm=False, distributed=False, **kwargs):
        '''
        Extracts the eval_trace value at the given time.
        Default: time=0 assuming this is the index for the NON-REVERSED trace. But the code will take it from the end since the input signal is TIME REVERSED.
        '''
        return self.eval_trace(inputs, pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs)[:,-(time+1),:].unsqueeze(1)                 # [batch_size, time_dim, x_dim]

    def forward(formula, inputs, pscale=1, scale=-1, keepdim=True, agm=False, distributed=False, **kwargs):
        '''
        Evaluates the robustness_trace given the input. The input is converted to the numerical value first.
        '''
        if isinstance(inputs, Expression):
            assert inputs.value is not None, "Input Expression does not have numerical values"
            return formula.robustness_trace(inputs.value, pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs)
        elif isinstance(inputs, torch.Tensor):
            return formula.robustness_trace(inputs, pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs)
        elif isinstance(inputs, tuple):
            return formula.robustness_trace(convert_to_input_values(inputs), pscale=pscale, scale=scale, keepdim=keepdim, agm=agm, distributed=distributed, **kwargs)
        else:
            raise ValueError("Not a invalid input trace")

    def __str__(self):
        raise NotImplementedError("__str__ not yet implemented")

    def __and__(phi, psi):
        return And(phi, psi)

    def __or__(phi, psi):
        return Or(phi, psi)

    def __invert__(phi):
        return Negation(phi)
        
class Expression(torch.nn.Module):
    '''
    Wraps a pytorch arithmetic operation, so that we can intercept and overload comparison operators.
    Expression allows us to express tensors using their names to make it easier to code up and read,
    but also keep track of their numeric values.
    '''
    def __init__(self, name, value):
        super(Expression, self).__init__()
        self.name = name
        self.value = value

    def set_name(self, new_name):
        self.name = new_name

    def set_value(self, new_value):
        self.value = new_value

    def __neg__(self):
        return Expression(-self.value)

    def __add__(self, other):
        if isinstance(other, Expression):
            return Expression(self.name + '+' + other.name, self.value + other.value)
        else:
            return Expression(self.name + "+other", self.value + other)

    def __radd__(self, other):
        return self.__add__(other)
        # No need for the case when "other" is an Expression, since that
        # case will be handled by the regular add

    def __sub__(self, other):
        if isinstance(other, Expression):
            return Expression(self.name + '-' + other.name, self.value - other.value)
        else:
            return Expression(self.name + "-other", self.value - other)


    def __rsub__(self, other):
        return Expression(other - self.value)
        # No need for the case when "other" is an Expression, since that
        # case will be handled by the regular sub

    def __mul__(self, other):
        if isinstance(other, Expression):
            return Expression(self.name + '*' + other.name, self.value * other.value)
        else:
            return Expression(self.name + "*other", self.value * other)


    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(a, b):
        # This is the new form required by Python 3
        numerator = a
        denominator = b
        num_name = 'num'
        denom_name = 'denom'
        if isinstance(numerator, Expression):
            num_name = numerator.name
            numerator = numerator.value
        if isinstance(denominator, Expression):
            denom_name = denominator.name
            denominator = denominator.value
        return Expression(num_name + '/' + denom_name, numerator/denominator)

    # Comparators
    def __lt__(lhs, rhs):
        assert isinstance(lhs, str) | isinstance(lhs, Expression), "LHS of LessThan needs to be a string or Expression"
        assert not isinstance(rhs, str), "RHS cannot be a string"
        return LessThan(lhs, rhs)

    def __le__(lhs, rhs):
        assert isinstance(lhs, str) | isinstance(lhs, Expression), "LHS of LessThan needs to be a string or Expression"
        assert not isinstance(rhs, str), "RHS cannot be a string"
        return LessThan(lhs, rhs)

    def __gt__(lhs, rhs):
        assert isinstance(lhs, str) | isinstance(lhs, Expression), "LHS of GreaterThan needs to be a string or Expression"
        assert not isinstance(rhs, str), "RHS cannot be a string"
        return GreaterThan(lhs, rhs)

    def __ge__(lhs, rhs):
        assert isinstance(lhs, str) | isinstance(lhs, Expression), "LHS of GreaterThan needs to be a string or Expression"
        assert not isinstance(rhs, str), "RHS cannot be a string"
        return GreaterThan(lhs, rhs)

    def __eq__(lhs, rhs):
        assert isinstance(lhs, str) | isinstance(lhs, Expression), "LHS of Equal needs to be a string or Expression"
        assert not isinstance(rhs, str), "RHS cannot be a string"
        return Equal(lhs, rhs)

    def __ne__(lhs, rhs):
        raise NotImplementedError("Not supported yet")

    def __str__(self):
        return str(self.name)

def make_stl_graph(form, node_attr=dict(style='filled',
                                          shape='box',
                                          align='left',
                                          fontsize='12',
                                          ranksep='0.1',
                                          height='0.2'),
                         graph_attr=dict(size="12,12")):
    """ Produces Graphviz representation of PyTorch autograd graph.
    Blue nodes are the Variables that require grad, orange are Tensors
    saved for backward in torch.autograd.Function
    Args:
        var: output Variable
        params: dict of (name, Variable) to add names to node that
            require grad (TODO: make optional)
    """

    # node_attr = dict(style='filled',
    #                  shape='box',
    #                  align='left',
    #                  fontsize='12',
    #                  ranksep='0.1',
    #                  height='0.2')
    dot = Digraph(node_attr=node_attr, graph_attr=graph_attr)
    seen = set()

    def size_to_str(size):
        return '(' + (', ').join(['%d' % v for v in size]) + ')'

    def tensor_to_str(tensor):
        device = tensor.device.type
        req_grad = tensor.requires_grad
        if req_grad == False:
            return "input"
        tensor = tensor.detach()
        if device == "cuda":
            tensor = tensor.cpu()
        return str(tensor.numpy())

    def add_nodes(form):
        # green are optimization variables
        # blue are non-optimization variables
        # orange are formula nodes
        # red is ambiguous, could be an optimization variable or it could not.

        if torch.is_tensor(form):
            color = "palegreen" if form.requires_grad else "lightskyblue"
            dot.node(str(id(form)), tensor_to_str(form), fillcolor=color)
        elif isinstance(form, Expression):
            color = "palegreen" if form.value.requires_grad else "lightskyblue"
            dot.node(str(id(form)), form.name, fillcolor=color)
        elif type(form) == str:
            dot.node(str(id(form)), form, fillcolor="lightcoral")
        elif isinstance(form, STL_Formula):
            dot.node(str(id(form)), form._get_name() + "\n" + str(form), fillcolor="orange")
        else:
            dot.node(str(id(form)), str(form), fillcolor="lightcoral")

        # recursive call to all the components of the formula
        if hasattr(form, '_next_function'):
            for u in form._next_function():
                dot.edge(str(id(u)), str(id(form)))
                add_nodes(u)

    # def add_nodes(var):
    #     if var not in seen:
    #         if torch.is_tensor(var):
    #             # note: this used to show .saved_tensors in pytorch0.2, but stopped
    #             # working as it was moved to ATen and Variable-Tensor merged
    #             dot.node(str(id(var)), size_to_str(var.size()), fillcolor='orange')
    #         elif hasattr(var, 'variable'):
    #             u = var.variable
    #             name = param_map[id(u)] if params is not None else ''
    #             node_name = '%s\n %s' % (name, size_to_str(u.size()))
    #             dot.node(str(id(var)), node_name, fillcolor='lightblue')
    #         elif var in output_nodes:
    #             dot.node(str(id(var)), str(type(var).__name__), fillcolor='darkolivegreen1')
    #         else:
    #             dot.node(str(id(var)), str(type(var).__name__))
    #         seen.add(var)
    #         if hasattr(var, 'next_functions'):
    #             for u in var.next_functions:
    #                 if u[0] is not None:
    #                     dot.edge(str(id(u[0])), str(id(var)))
    #                     add_nodes(u[0])
    #         if hasattr(var, 'saved_tensors'):
    #             for t in var.saved_tensors:
    #                 dot.edge(str(id(t)), str(id(var)))
    #                 add_nodes(t)

    
    # handle multiple outputs
    if isinstance(form, tuple):
        for v in form:
            add_nodes(v)
    else:
        add_nodes(form)

    resize_graph(dot)

    return dot


def resize_graph(dot, size_per_element=0.15, min_size=12):
    """Resize the graph according to how much content it contains.
    Modify the graph in place.
    """
    # Get the approximate number of nodes and edges
    num_rows = len(dot.body)
    content_size = num_rows * size_per_element
    size = max(min_size, content_size)
    size_str = str(size) + "," + str(size)
    dot.graph_attr.update(size=size_str)


def save_graph(dot, filename, format='pdf', cleanup=True):
    dot.render(filename=filename, format=format, cleanup=cleanup)